function [ gmp ] = ghostMap( imSet, expT )
%GHOSTMAP Extract the potential ghost map from multiple exposures
% 
%   output = ghostMap(imSet, expT) returns the ghost map for a set of
%   images whose exposure times are given in expT. 
%
%   INPUTS:
%   imSet - Cell array of n-by-m-by-3 input images. It is assumed that the
%   images have linear camera response function.
%   expT  - Vector of image exposure times. expT(i) must be equal to the
%   exposure time of imSet{i}.
%
%   OUTPUTS:
%   gmp   - n-by-m bitmap of potential ghost regions. 
%
%   SAMPLE RUN:
%   imSet{1} = imread(fullfile(pwd, 'sample_inputs', '4.tiff'));
%   imSet{2} = imread(fullfile(pwd, 'sample_inputs', '5.tiff'));
%   imSet{3} = imread(fullfile(pwd, 'sample_inputs', '6.tiff'));
%   expT = [1/512 1/256 1/128];
%   gmp = ghostMap(imSet, expT);
%   figure, imshow(gmp);

%   Version History:
%   6 Dec 2015 - Initial release.
%
%   Copyright (c) 2015, Okan Tarhan Tursun.


gmp = false(size(imSet{1},1), size(imSet{1},2));

se = strel('disk', 2);

filt = fspecial('gaussian', 9,1);

for n=1:numel(imSet)-1
    for m=1:size(imSet{n},3)
        ratio = expT(n+1) / expT(n);
        if ratio > 1
            im1 = im2double(imSet{n});
            im2 = im2double(imSet{n+1});
        else
            im1 = im2double(imSet{n+1});
            im2 = im2double(imSet{n});
            ratio = 1 ./ ratio;
        end
        im1Rad = imfilter(double(im1(:,:,m)),filt)*ratio;
        im2Rad = imfilter(double(im2(:,:,m)), filt);
        im1Rad(im1Rad > max(im2Rad(:))) = max(im2Rad(:)); % clip the first image
         
        tmp = im1Rad ./ im2Rad - 1;
        tmp = abs(tmp);
 
        avg1 = mean(im1*ratio,3);
        avg2 = mean(im2,3);
        avg = mean(cat(3, avg1, avg2), 3);
        tmp = tmp >= (2*avg-1).^12+0.1;

        tmp = imopen(tmp, se);
        
        gmp = gmp | tmp;
    end
end
