function [ hdrIm, hdrImBE, hdrIm_refPatched, hdrImLap, gmp ] = hdrMerge( imSet, expT )
%HDRMERGE Simple HDR merging of input LDR images with linear camera response function
%
%   [hdrIm, hdrImBE, hdrIm_refPatched, hdrImLap, gmp] = hdrMerge(imSet,
%   expT) returns HDR images reconstructed from given set of LDR images
%   with knwon exposure times.
%
%   INPUTS:
%   imSet - Cell array of n-by-m-by-3 input images. It is assumed that the
%   images have linear camera response function.
%   expT  - Vector of image exposure times. expT(i) must be equal to the
%   exposure time of imSet{i}.
%
%   OUTPUTS:
%   hdrIm            - HDR image reconstructed without any deghosting.
%   hdrImBE          - HDR image reconstructed using the best exposure for
%   each pixel. 
%   hdrIm_refPatched - HDR image reconstructed using the reference (middle)
%   exposure in motion regions.
%   hdrImLap         - HDR image reconstructed using the reference (middle)
%   exposure and Laplacian blending. This is the output used in our
%   work [1].
%   gmp              - Ghost map.
%
%   SAMPLE RUN:
%   imSet{1} = imread(fullfile(pwd, 'sample_inputs', '4.tiff'));
%   imSet{2} = imread(fullfile(pwd, 'sample_inputs', '5.tiff'));
%   imSet{3} = imread(fullfile(pwd, 'sample_inputs', '6.tiff'));
%   expT = [1/512 1/256 1/128];
%   [hdrIm, hdrImBE, hdrIm_refPatched, hdrImLap, gmp] = hdrMerge(imSet, expT);
%   figure, imshow(tonemap(hdrIm));
%   title('HDR without deghosting');
%   figure, imshow(tonemap(hdrImBE));
%   title('HDR using the best exposure');
%   figure, imshow(tonemap(hdrIm_refPatched));
%   title('HDR by patching using the reference (middle) exposure - no blending');
%   figure, imshow(tonemap(hdrImLap));
%   title('HDR by patching using the reference (middle) exposure - Laplacian blending');
%
%   [1] Tursun, O. T., Akyuz, A. O., Erdem, A., & Erdem, E. (2015). The State
%   of the Art in HDR Deghosting: A Survey and Evaluation. In Computer
%   Graphics Forum (Vol. 32, pp. 348-362).

%   Version History:
%   6 Dec 2015 - Initial release.
%
%  Copyright (c) 2015, Okan Tarhan Tursun
%



addpath(fullfile(pwd, 'image_pyramid'));

initialWeightMap = cell(size(imSet));
radianceMap = cell(numel(imSet),1);

%% Calculate initial weight maps
h = waitbar(0,'Calculating Initial Weight Maps');
for n=1:numel(imSet)
    imSet{n} = im2double(imSet{n});
    for ch=1:size(imSet{1},3)
        intensity = imSet{n}(:,:,ch); % get channel intensity
        initialWeightMap{n}(:,:,ch) = triangular_weight_func(intensity);
        radiance = intensity / expT(n);
        radianceMap{n}(:,:,ch) = radiance;
        waitbar(((n-1)*size(imSet{1},3)+ch) / (numel(imSet) * size(imSet{1},3)));
    end
end
close(h);

middleExpIdx = floor((numel(imSet)-1) / 2)+1;
weightMap = zeros([size(imSet{1},1) size(imSet{1},2) numel(imSet)]);
for n=1:numel(imSet)
    % weightMap of each image is the average of weight maps used for
    % different color channels
    weightMap(:,:,n) = mean(initialWeightMap{n},3);
end
zeroWeightMask = sum(weightMap,3) == 0;
tmp = weightMap(:,:,middleExpIdx);
tmp(zeroWeightMask) = 1;
weightMap(:,:,middleExpIdx) = tmp; % if all the input images have zero weight for a particular pixel, use the middle exposed image

normWeightMap = weightMap ./ repmat(sum(weightMap,3), [1 1 numel(imSet)]);

% set the weight of the best exposed image to unity
weightMapBE = double(abs(weightMap - repmat(max(weightMap,[],3), [1 1 size(weightMap,3)])) < 1E-9);

%% Construct the HDR image w/o deghosting
hdrIm = 0;
for k=1:numel(imSet)
    hdrIm = hdrIm + repmat(normWeightMap(:,:,k), [1 1 3]) .* radianceMap{k};
end

%% Construct HDR image using only the best exposed image in the ghost regions (no blending)
hdrImBE = 0;
for k=1:numel(imSet)
    hdrImBE = hdrImBE + repmat(weightMapBE(:,:,k), [1 1 3]) .* radianceMap{k};
end
hdrImBE = hdrImBE ./ repmat(sum(weightMapBE,3), [1 1 3]);
hdrImBE(isnan(hdrImBE(:))) = 0;

%% Remove ghosts by patching the ghost regions using the reference image (no blending)
gmp_orig = ghostMap(imSet, expT);
se = strel('disk', 2);
gmp = imdilate(gmp_orig, se);

hdrIm_refPatched = hdrIm; % hdrImg patched using only reference image

refImIdx = floor((numel(imSet)-1) / 2) + 1;
for i=1:size(hdrIm,3)
    tmp = hdrIm_refPatched(:,:,i);
    tmp_refIm = radianceMap{refImIdx}(:,:,i);
    tmp(gmp) = tmp_refIm(gmp);
    hdrIm_refPatched(:,:,i) = tmp;
end

%% Remove ghosts by patching the ghost regions using the reference image and Laplacian blending
level = 3;

limga = genPyr(hdrIm,'lap',level); % the Laplacian pyramid
refRad = im2double(imSet{refImIdx}) / expT(refImIdx);
limgb = genPyr(refRad,'lap',level);

maska = double(~gmp);

limgo = cell(1,level); % the blended pyramid
for p = 1:level
	[Mp, Np, ~] = size(limga{p});
	maskap = imresize(maska,[Mp Np]);
	maskbp = 1-maskap; 
	limgo{p} = limga{p}.*repmat(maskap, [1 1 size(limga{p},3)]) + limgb{p}.*repmat(maskbp, [1 1 size(limgb{p},3)]);
end
hdrImLap = pyrReconstruct(limgo);
hdrImLap(hdrImLap<0) = 0;

padSize = [size(hdrIm,1) - size(hdrImLap,1), size(hdrIm,2) - size(hdrImLap,2) ];
hdrImLap = padarray(hdrImLap, padSize, 'post');
for c = 1:size(hdrImLap,3)
    hdrImLap(end-padSize(1):end, :, :) =  hdrIm(end-padSize(1):end, :, :);
    hdrImLap(:, end-padSize(2):end, :) =  hdrIm(:, end-padSize(2):end, :);
end

end


function w = triangular_weight_func(x)
    w = 1 - abs(2*(x-0.5));
end

function w = broadhat_weight_func(x)
	w = 1 - (2*x - 1).^12;
end

function w = gaussian_weight_func(x)
	w = exp( -1.0 * 25.0 * (x-0.5) .^ 2.0 );
end
